from importlib import import_module
import asyncio
from typing import Type
from arkitekt import Arkitekt
from watchfiles import awatch
from watchfiles.filters import BaseFilter, PythonFilter
import os





def import_builder(builder) -> Type[Arkitekt]:
    module_path, function_name = builder.rsplit(".", 1)
    module = import_module(module_path)
    function = getattr(module, function_name)
    return function



async def build_and_run(builder, identifier, version, entrypoint):

    app = builder(identifier, version)

    import_module(entrypoint)

    async with app:
        await app.rekuest.run()


    

class EntrypointFilter(PythonFilter):

    def __init__(self, entrypoint, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.entrypoint = entrypoint


    def __call__(self, change, path: str) -> bool:
        x = super().__call__(change, path)
        if not x:
            return False

        x = os.path.basename(path)
        module_name = x.split(".")[0]

        return module_name == self.entrypoint



async def dev_module(identifier, version, entrypoint,  builder: str = "arkitekt.builders.easy"):


    builder_func = import_builder(builder)
    x = asyncio.create_task(build_and_run(builder_func, identifier, version, entrypoint))



    async for changes in awatch(".", watch_filter=EntrypointFilter(entrypoint), debounce=2000):
        
        print("Canceling")

        if x.done():
            print(x.exception())
            print("Restarting")

        else:
            x.cancel()
            try:
                await x;
            except asyncio.CancelledError:
                pass

        print("Restarting")
        x = asyncio.create_task(build_and_run(builder_func, identifier, version, entrypoint))


   

    
    










from .herre import HerreApp
from unlok import Unlok
from unlok.rath import UnlokRath, UnlokLinkComposition
from pydantic import Field
from rath.links.split import SplitLink
from rath.contrib.fakts.links.aiohttp import FaktsAIOHttpLink
from rath.contrib.fakts.links.websocket import FaktsWebsocketLink
from rath.contrib.herre.links.auth import HerreAuthLink
from mikro.contrib.fakts.datalayer import FaktsDataLayer
from graphql import OperationType


class ArkitektUnlok(Unlok):
    rath: UnlokRath = Field(
        default_factory=lambda: UnlokRath(
            link=UnlokLinkComposition(
                auth=HerreAuthLink(),
                split=SplitLink(
                    left=FaktsAIOHttpLink(fakts_group="herre"),
                    right=FaktsWebsocketLink(fakts_group="herre"),
                    split=lambda o: o.node.operation != OperationType.SUBSCRIPTION,
                ),
            )
        )
    )


class UnlokApp(HerreApp):
    """Lok App"""

    lok: ArkitektUnlok = Field(default_factory=ArkitektUnlok)



import rich_click as click
import asyncio
from arkitekt.cli.init import Manifest, load_manifest, write_manifest
import subprocess
from rich.console import Console
from rich.progress import Progress, TextColumn
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel
import os
import sys



console = Console()

logo = """
            _    _ _       _    _   
  __ _ _ __| | _(_) |_ ___| | _| |_ 
 / _` | '__| |/ / | __/ _ \ |/ / __|
| (_| | |  |   <| | ||  __/   <| |_ 
 \__,_|_|  |_|\_\_|\__\___|_|\_\\__|
                                                                   
"""

welcome = """
Welcome to Arkitekt. Arkitekt is a framework for building beautiful and fast (serverless) APIs around
your python code.
To get started, we need some information about your app. You can always change this later.
"""


template_app = '''
from arkitekt import register

@register()
def test_func(a: int, b: int) -> int:
    """Add two numbers

    This function adds two numbers and returns the result.

    Args:
        a (int): The first number
        b (int): The second number

    Returns:
        int: The returned number
    """
    return a + b

'''


default_docker_file = '''
FROM python:3.8-slim-buster


RUN pip install arkitekt==0.4.23


RUN mkdir /app
COPY . /app
WORKDIR /app

'''




click.rich_click.HEADER_TEXT = logo
click.rich_click.ERRORS_EPILOGUE = "To find out more, visit [link=https://jhnnsrs.github.io/doks]https://jhnnsrs.github.io/doks[/link]"
click.rich_click.USE_RICH_MARKUP = True


def compile_scopes():

    return ["read", "write"]


def compile_builders():

    return ["arkitekt.builders.easy", "arkitekt.builders.port"]

def compile_runtimes():

    return ["nvidia", "standard"]

@click.group()
@click.pass_context
def cli(ctx):
    """Arkitekt is a framework for building beautiful and fast (serverless) APIs around
    your python code. 
    It is build on top of Rekuest and is designed to be easy to use."""

    sys.path.append(os.getcwd())
    pass



@cli.group()
def run():
    """ Runs the arkitekt app (using a builder)"""
    pass

@run.command()
@click.option('--url', help='The fakts url for connection', default="http://localhost:8000/f/")
@click.option('--public_url', help='The fakts public_url for connection', default="http://localhost:8000/f/")
@click.option('--instance_id',"-i", help='The fakts instance_id for connection', default="main")
def easy(url, public_url, instance_id):
    """Runs the arkitekt app using the easy builder, which is the default builder
    
    This builder is the default builder for all script based apps. It is designed to be easy to use and
    to get started with. It is not recommended to use this builder for production apps.
    """

    from arkitekt.cli.run import run_costum, run_easy, run_port
    manifest = load_manifest()
    if not manifest:
        raise click.ClickException("No manifest found. Please run `arkitekt init` first before deploying an arkitekt app.")

    asyncio.run(run_easy(manifest.entrypoint, manifest.identifier, manifest.version, url, public_url, instance_id))

@run.command()
@click.option('--url', help='The fakts url for connection')
@click.option('--token', help='The fakts token for connection')
def port(url, token):
    """Runs the arkitekt app"""

    from arkitekt.cli.run import run_costum, run_easy, run_port
    manifest = load_manifest()
    if not manifest:
        raise click.ClickException("No manifest found. Please run `arkitekt init` first before deploying an arkitekt app.")

    asyncio.run(run_port(manifest.entrypoint, manifest.identifier, manifest.version, url, token ))

@run.command()
@click.option('--builder', help='The builder used to construct the app', type=str, default="arkitekt.builders.easy")
def custom(builder):
    """Runs the arkitekt app"""


    from arkitekt.cli.run import run_costum, run_easy, run_port
    manifest = load_manifest()
    if not manifest:
        raise click.ClickException("No manifest found. Please run `arkitekt init` first before deploying an arkitekt app.")

    asyncio.run(run_costum(manifest.entrypoint, manifest.identifier, manifest.version, builder=builder))



@cli.command()
@click.option('--builder', help='The builder used to construct the app', type=str, default="arkitekt.builders.easy")
def dev(builder):
    """Runs the arkitekt app"""

    from arkitekt.cli.dev import dev_module
    manifest = load_manifest()
    if not manifest:
        raise click.ClickException("No manifest found. Please run `arkitekt init` first before deploying an arkitekt app.")

    asyncio.run(dev_module(manifest.identifier, manifest.version, manifest.entrypoint, builder=builder))


@cli.command()
@click.option('--app', prompt='The module path', help='The module path', default="app")
def scan(app):

    from arkitekt.cli.scan import scan_module
    """Scans your arkitekt app for leaking variables"""
    variables = scan_module(app)

    if not variables:
        click.echo("No dangerous variables found. You are good to go!")
        return

    for key, value in variables.items():
        click.echo(f"{key}: {value}")

@cli.group()
def deploy():
    """ Deploys the arkitekt app to a specific platform (like port)"""
    pass


def search_username_in_docker_info(docker_info: str):
    for line in docker_info.splitlines():
        if "Username" in line:
            return line.split(":")[1].strip()


@deploy.command()
@click.option('--version', help='The version of your app')
@click.option('--dockerfile', help='The dockerfile to use')
@click.option('--nopush', help='Skip push')
@click.option('--tag', help='The tag to use')
@click.option('--runtime', help='The runtime to use', type=click.Choice(compile_runtimes()))
@click.option('--nodefs', help='Do not inspect definitions', is_flag=True)
@click.option('--builder', help='The builder to use', type=str, default="arkitekt.builder.port")
@click.option('--entrypoint', help='The entrypot to use', type=str)
def port(version, dockerfile, nopush, tag, runtime, nodefs, builder, entrypoint):
    """Deploys the arkitekt app to port"""

    from arkitekt.cli.deploy import generate_deployment

    manifest = load_manifest()
    if not manifest:
        raise click.ClickException("No manifest found. Please run `arkitekt init` first before deploying an arkitekt app.")


    entrypoint = entrypoint or manifest.entrypoint

    if not os.path.exists(f"Dockerfile"):
        if click.confirm('Dockerfile does not exists. Do you want to generate a template Dockerfile?'):
            with open(f"Dockerfile", "w") as f:
                f.write(default_docker_file)    
                console.print("Dockerfile generated. Please edit it to your needs. And restart the deployment.")
                return 


    md = Panel("Deploying to Port", subtitle="This may take a while...", subtitle_align="right")
    console.print(md)
    version = version or manifest.version
    identifier = manifest.identifier
    entrypoint = entrypoint or manifest.entrypoint

    if version == "dev":
        raise click.ClickException("You cannot deploy a dev version. Please change the version in your manifest. Or use the --version flag.")

    with console.status("Searching username"):
        docker_info = subprocess.check_output(["docker", "info"]).decode("utf-8")
        username = search_username_in_docker_info(docker_info)
        if not username:
            raise click.ClickException("Could not find username in docker info. Have you logged in? Try 'docker login'")


    tag = tag or click.prompt("The tag to use", default=f"{username}/{manifest.identifier}:{version}")     
    runtime = runtime or click.prompt("The runtime to use", type=click.Choice(compile_runtimes()), default="standard")


    deployed = {
        "runtime": runtime,
    }

    md = Panel("Building Docker Container")
    console.print(md)

    docker_run = subprocess.run(["docker", "build", "-t", tag, "-f", dockerfile or "Dockerfile", "."])
    if docker_run.returncode != 0:
        raise click.ClickException("Could not build docker container")

    if not nopush:
        md = Panel("Pushing Docker Container")
        console.print(md)
        docker_run = subprocess.run(["docker", "push", tag])
        if docker_run.returncode != 0:
            raise click.ClickException("Could not push docker container")



    deployed["docker"] = tag

    generate_deployment(identifier, version, entrypoint, deployed, "arkitekt.deployers.port.dockerbuild", manifest.scopes, with_definitions=not nodefs, builder=builder)


def prompt_scopes():
    used_scopes = []
    scopes = compile_scopes()

    while click.confirm("Do you want to add a scope?"):
        scope = click.prompt("The scope to use", type=click.Choice(scopes), show_choices=True)
        used_scopes.append(scope)

    return used_scopes



@cli.command()
@click.option('--identifier', help='The identifier of your app')
@click.option('--version', help='The version of your app')
@click.option('--entrypoint', help='The version of your app')
@click.option('--author', help='The author of your app')
@click.option('--builder', help='The builder used to construct the app')
@click.option('--scopes', "-s", help='The scope of the app', type=click.Choice(compile_scopes()), multiple=True)
def init(identifier, version, author, entrypoint, builder, scopes):
    """Initializes the arkitekt app"""

    md = Panel(logo + welcome, title="Welcome to Arkitekt", title_align="center")
    console.print(md)

    oldmanifest = load_manifest()

    if oldmanifest:
        if not click.confirm('Do you want to overwrite your old configuration'):
            return



    manifest = Manifest(
            identifier=identifier or click.prompt("Your apps identifier", default=getattr(oldmanifest, "identifier", os.path.basename(os.getcwd()))),
            version=version or click.prompt("The apps version", default=getattr(oldmanifest, "version", "dev")),
            entrypoint=entrypoint or click.prompt("The module path", default=getattr(oldmanifest, "entrypoint", "app")),
            author=author or click.prompt("The apps author", default=getattr(oldmanifest, "author", "john doe")),
            builder=builder or click.prompt("The builder to use", default=getattr(oldmanifest, "builder", "arkitekt.builders.easy")),
            scopes=scopes or prompt_scopes()
        )

    entrypoint = manifest.entrypoint

    if not os.path.exists(f"{entrypoint}.py"):
        if click.confirm('Entrypoint does not exists. Do you want to generate a python file?'):
            with open(f"{entrypoint}.py", "w") as f:
                f.write(template_app)


    write_manifest(manifest)


@cli.command()
def version():
    """Shows the version of arkitekt"""
    click.echo(f"Arkitekt {__version__}")



if __name__ == '__main__':
    cli()
   


from importlib import import_module
from arkitekt.utils import create_arkitekt_folder
from pydantic import BaseModel, Field
import os
from typing import List, Union, Optional
import yaml
import json
import datetime

def import_deployer(builder):
    module_path, function_name = builder.rsplit(".", 1)
    module = import_module(module_path)
    function = getattr(module, function_name)
    return function

def generate_definitions(module_path: str = "hu"):
    from rekuest.definition.registry import get_default_definition_registry

    module_path = f"hu"
    module = import_module(module_path)

    reg = get_default_definition_registry()
    return list(reg.definitions.keys())



class Manifest(BaseModel):
    identifier: str
    version: str
    author: str
    entrypoint: str
    scopes: List[str]
    created_at: datetime.datetime = Field(default_factory=datetime.datetime.now)

def load_manifest() -> Optional[Manifest]:
    path = create_arkitekt_folder()
    config_file = os.path.join(path, f"manifest.yaml")
    if os.path.exists(config_file):
        with open(config_file, "r") as file:
            manifest = yaml.safe_load(file)
        return Manifest(**manifest)
    return None

def write_manifest(manifest: Manifest):
   
    path = create_arkitekt_folder()
    config_file = os.path.join(path, f"manifest.yaml")

    with open(config_file, "w") as file:
        documents = yaml.safe_dump(json.loads(manifest.json(exclude_none=True, exclude_unset=True)), file, sort_keys=True)







from fakts import FaktsGrant
from fakts.fakts import Fakts
from fakts.grants.remote.base import RemoteGrant
from fakts.grants.remote.device_code import DeviceCodeGrant
from koil.composition.base import Composition
from pydantic import Field


class ArkitektFakts(Fakts):
    grant: FaktsGrant = Field(default_factory=DeviceCodeGrant)


class FaktsApp(Composition):
    fakts: ArkitektFakts = Field(default_factory=ArkitektFakts)
    """The fakts layer that is used for configuration
    """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['arkitekt',
 'arkitekt.apps',
 'arkitekt.cli',
 'arkitekt.deployers',
 'arkitekt.qt']

package_data = \
{'': ['*'], 'arkitekt.qt': ['assets/dark/*', 'assets/light/*']}

install_requires = \
['fakts>=0.3.11',
 'fluss>=0.1.38',
 'herre>=0.3.10',
 'mikro>=0.3.37',
 'rekuest>=0.1.25',
 'unlok>=0.1.10']

extras_require = \
{'cli': ['rich-click>=1.6.1,<2.0.0', 'watchfiles>=0.18.1,<0.19.0']}

entry_points = \
{'console_scripts': ['arkitekt = arkitekt.cli.main:cli']}

setup_kwargs = {
    'name': 'arkitekt',
    'version': '0.4.43',
    'description': 'client for the arkitekt platform',
    'long_description': 'None',
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from arkitekt.apps.herre import HerreApp
from pydantic import Field
from fluss.fluss import Fluss
from fluss.rath import FlussLinkComposition, FlussRath
from rath.links.split import SplitLink
from rath.contrib.fakts.links.aiohttp import FaktsAIOHttpLink
from rath.contrib.fakts.links.websocket import FaktsWebsocketLink
from rath.contrib.herre.links.auth import HerreAuthLink
from graphql import OperationType
from .connected import ConnectedApp
from .rekuest import RekuestApp
from .unlok import UnlokApp
from arkitekt.healthz import FaktsHealthz


class ArkitektFluss(Fluss):
    rath: FlussRath = Field(
        default_factory=lambda: FlussRath(
            link=FlussLinkComposition(
                auth=HerreAuthLink(),
                split=SplitLink(
                    left=FaktsAIOHttpLink(fakts_group="fluss"),
                    right=FaktsWebsocketLink(fakts_group="fluss"),
                    split=lambda o: o.node.operation != OperationType.SUBSCRIPTION,
                ),
            )
        )
    )
    healthz: FaktsHealthz = Field(
        default_factory=lambda: FaktsHealthz(fakts_group="fluss")
    )


class FlussApp(HerreApp):
    """Mikro App

    It is responsible for setting up the connection to the mikro-server and
    handling authentification and setting up the configuration. Mikro handles the creation of the datalayer and
    the graphql client.

    You can compose this app together with other apps to create a full fledged app. (Like combining with
    arkitekt to enable to call functions that you define on the app). See the example in the docstring.

    Attributes:
        fakts (Fakts): The fakts instance to use.
        mikro (Mikro): The mikro instance to use.
        herre (Herre): The herre instance to use.

    """

    fluss: ArkitektFluss = Field(default_factory=ArkitektFluss)


class ConnectedFluss(RekuestApp, UnlokApp, FlussApp):
    pass



from importlib import import_module
from typing import Type
from arkitekt import Arkitekt





module_path = f"hu"

z = locals()
y = locals()


def import_builder(module_path, function_name) -> Type[Arkitekt]:
    module = import_module(module_path)
    function = getattr(module, function_name)
    return function


async def run_app(entrypoint, app):
    try:
        import_module(entrypoint)
    except ModuleNotFoundError as e:
        print(f"Could not find entrypoint module {entrypoint}")
        raise e

    async with app:
        await app.rekuest.run()

async def run_costum(entrypoint, identifier, version,   builder: str = "arkitekt.builders.easy"):

    module_path, function_name = builder.rsplit(".", 1)
    builder = import_builder(module_path, function_name)



    app = builder(identifier, version)

    await run_app(entrypoint, app)



async def run_easy(entrypoint, identifier, version, url, public_url, instance_id):
    from arkitekt.builders import easy

    app = easy(identifier, version, url, instance_id=instance_id)


    await run_app(entrypoint, app)



async def run_port(entrypoint, identifier, version, url, token):
    from arkitekt.builders import port

    app = port(identifier, version, url=url, token=token)


    await run_app(entrypoint, app)    
    










from arkitekt.apps.fakts import FaktsApp
from herre.grants.fakts import FaktsGrant
from herre.herre import Herre
from pydantic import Field


class ArkitektHerre(Herre):
    grant: FaktsGrant = Field(default_factory=lambda: FaktsGrant())



class HerreApp(FaktsApp):
    herre: ArkitektHerre = Field(default_factory=ArkitektHerre)
    """The fakts layer thats is used for configuration
    """



from arkitekt.apps.mikro import MikroApp
from arkitekt.apps.rekuest import RekuestApp
from arkitekt.apps.unlok import UnlokApp
from typing import Optional


class ConnectedApp(MikroApp, RekuestApp, UnlokApp):
    identifier: str
    version: str
    pass

    def _repr_html_(self):
        return (
            f'''<div><h4>{self.identifier or 'Arkitekt App'} {f"@{self.version}" if self.version else ""}</h4><table>'''
            + "\n".join(["<tr><td>{}</td><td>{}</td></tr>".format(key, value._repr_html_inline_()) for key, value in self if hasattr(value, "_repr_html_inline_")])
            + "</table></div>"
        )


class App(ConnectedApp):
    pass
